﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows.Data;
using System.Windows.Controls;
using System.Globalization;
using System.ComponentModel;
using System.Collections.ObjectModel;
using System.Windows.Markup;
using System.Reflection;
using System.ComponentModel.DataAnnotations;
using System.Windows;

namespace WoVS.Presentation
{
    /// <summary>
    /// <see cref="Binding"/> decorator that adds transparent Enterprise Library 
    /// validation.
    /// </summary>
    public class ValidationBinding : MarkupExtension
    {
        Binding binding;
        BindingExpression expression;

        /// <summary>
        /// Default Constructor
        /// </summary>
        public ValidationBinding()
        {
            binding = new Binding { ValidatesOnDataErrors = true, ValidatesOnExceptions = true };
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="ValidationBinding"></see> class with an initial path.
        /// </summary>
        /// <param name="path">The initial <see cref="Binding.Path"></see> for the binding.</param>
        public ValidationBinding(string path)
        {
            binding = new Binding(path) { ValidatesOnDataErrors = true, ValidatesOnExceptions = true };
        }

        public ValidationStep ValidationStep { get; set; }

        /// <summary>
        /// See <see cref="BindingBase.ProvideValue"/>.
        /// </summary>
        public override object ProvideValue(IServiceProvider serviceProvider)
        {
            expression = binding.ProvideValue(serviceProvider) as BindingExpression;

            var target = (IProvideValueTarget)serviceProvider.GetService(typeof(IProvideValueTarget));
            var validator = new ValidatorRule(binding, expression) { ValidationStep = ValidationStep };

            DependencyPropertyChangedEventHandler refreshValidator = (sender, args) => validator.RefreshValidator();

            var fe = target as FrameworkElement;
            var fce = target as FrameworkContentElement;

            if (fe != null)
                fe.DataContextChanged += refreshValidator;
            else if (fce != null)
                fce.DataContextChanged += refreshValidator;

            binding.ValidationRules.Add(validator);

            return expression;
        }

        class ValidatorRule : ValidationRule
        {
            Binding binding;
            BindingExpression expression;
            Validator validator = new NullValidator();
            object currentValue;
            object dataItem;
            PropertyInfo targetProperty;

            public ValidatorRule(Binding binding, BindingExpression expression)
            {
                this.binding = binding;
                this.expression = expression;
            }

            public override System.Windows.Controls.ValidationResult Validate(object value, CultureInfo cultureInfo)
            {
                currentValue = value;

                // If the data item changed, reset validator.
                if (dataItem != expression.DataItem)
                    RefreshValidator();

                if (binding.Converter != null)
                {
                    try
                    {
                        currentValue = binding.Converter.ConvertBack(value, targetProperty.PropertyType, binding.ConverterParameter, cultureInfo);
                    }
                    catch (Exception ex)
                    {
                        // If we fail to convert, don't even get to the validators.
                        return new System.Windows.Controls.ValidationResult(false, ex.Message);
                    }
                }

                // TODO: support for value types and nullable types?
                if (currentValue != null && !targetProperty.PropertyType.IsAssignableFrom(currentValue.GetType()))
                {
                    TypeConverter converter = null;
                    if (Attribute.IsDefined(targetProperty, typeof(TypeConverterAttribute)))
                    {
                        string converterTypeName = ((TypeConverterAttribute)
                            Attribute.GetCustomAttribute(targetProperty, typeof(TypeConverterAttribute))).ConverterTypeName;
                        Type converterType = Type.GetType(converterTypeName);

                        if (converterType == null)
                        {
                            //ValidationBinding.TraceSource.TraceEvent(TraceEventType.Warning, 0,
                            //    Properties.Resources.CantLoadTypeConverterType,
                            //    converterTypeName,
                            //    targetProperty.DeclaringType.Name,
                            //    targetProperty.Name);
                        }
                        else
                        {
                            try
                            {
                                converter = (TypeConverter)Activator.CreateInstance(converterType);
                            }
                            catch
                            {
                                //ValidationBinding.TraceSource.TraceEvent(TraceEventType.Warning, 0,
                                //    Properties.Resources.CantLoadTypeConverterType,
                                //    converterTypeName,
                                //    targetProperty.DeclaringType.Name,
                                //    targetProperty.Name);
                            }
                        }
                    }

                    if (converter == null)
                    {
                        // It may be null if we couldn't load a custom converter, so 
                        // we fallback to the type-assigned one.
                        converter = TypeDescriptor.GetConverter(targetProperty.PropertyType);
                    }

                    if (!converter.CanConvertFrom(currentValue.GetType()))
                    {
                        return new System.Windows.Controls.ValidationResult(false, "Can't convert");
                        //return new System.Windows.Controls.ValidationResult(false, String.Format(
                        //    Properties.Resources.IncompatibleValueType,
                        //    currentValue,
                        //    targetProperty.DeclaringType.Name,
                        //    targetProperty.Name));
                    }
                    else
                    {
                        try
                        {
                            currentValue = converter.ConvertFrom(currentValue);
                        }
                        catch (Exception ex)
                        {
                            // If we fail to convert, don't even get to the validators.
                            return new System.Windows.Controls.ValidationResult(false, ex.Message);
                        }
                    }
                }

                return validator.Validate(value);
            }

            internal void RefreshValidator()
            {
                InitializeValidator();
                dataItem = expression.DataItem;
            }

            private void InitializeValidator()
            {
                if (expression.DataItem == null)
                {
                    targetProperty = null;
                    validator = new NullValidator();
                    return;
                }

                Type targetType = expression.DataItem.GetType();
                string path = binding.Path.Path;
                PropertyInfo property = targetType.GetProperty(path);

                // Retrieve target property from nested path (i.e. Customer.Order)
                int dot = path.IndexOf('.');
                while (dot >= 0)
                {
                    property = targetType.GetProperty(path.Substring(0, dot));

                    if (property == null)
                    {
                        targetProperty = null;
                        validator = new NullValidator();
                        return;
                    }

                    targetType = property.PropertyType;
                    path = path.Substring(dot + 1);
                    property = targetType.GetProperty(path);
                    dot = path.IndexOf('.');
                }

                if (property == null)
                {
                    targetProperty = null;
                    validator = new NullValidator();
                    return;
                }

                targetProperty = property;
                validator = new PropertyValidator(targetProperty
                    .GetCustomAttributes(typeof(ValidationAttribute), true)
                    .Cast<ValidationAttribute>()
                    .ToList(),
                    targetProperty.Name);
            }

            private object GetValue()
            {
                return currentValue;
            }

            abstract class Validator
            {
                public abstract ValidationResult Validate(object value);
            }

            class PropertyValidator : Validator
            {
                IEnumerable<ValidationAttribute> attributes;
                string propertyName;

                public PropertyValidator(IEnumerable<ValidationAttribute> attributes, string propertyName)
                {
                    this.attributes = attributes;
                    this.propertyName = propertyName;
                }

                public override ValidationResult Validate(object value)
                {
                    var errors = new StringBuilder();

                    foreach (var attribute in attributes)
                    {
                        if (!attribute.IsValid(value))
                            errors.AppendLine(attribute.FormatErrorMessage(propertyName));
                    }

                    return new ValidationResult(errors.Length == 0, errors.ToString());
                }
            }

            class NullValidator : Validator
            {
                public override ValidationResult Validate(object value)
                {
                    return new ValidationResult(true, "");
                }
            }
        }

        #region BindingBase

        /// <summary>Returns a value that indicates whether serialization processes should serialize the effective value of the <see cref="P:System.Windows.Data.BindingBase.FallbackValue"></see> property on instances of this class.</summary>
        /// <returns>true if the <see cref="P:System.Windows.Data.BindingBase.FallbackValue"></see> property value should be serialized; otherwise, false.</returns>
        [EditorBrowsable(EditorBrowsableState.Never)]
        public bool ShouldSerializeFallbackValue()
        {
            return (binding.ShouldSerializeFallbackValue());
        }

        /// <summary>Gets or sets the value to use when the binding is unable to return a value.</summary>
        /// <returns>The default value is <see cref="F:System.Windows.DependencyProperty.UnsetValue"></see>.</returns>
        public object FallbackValue
        {
            get { return binding.FallbackValue; }
            set { binding.FallbackValue = value; }
        }

        #endregion

        #region Binding

        /// <summary>Indicates whether the <see cref="P:System.Windows.Data.Binding.Path"></see> property should be persisted.</summary>
        /// <returns>true if the property value has changed from its default; otherwise, false.</returns>
        [EditorBrowsable(EditorBrowsableState.Never)]
        public bool ShouldSerializePath()
        {
            return binding.ShouldSerializePath();
        }
        /// <summary>Indicates whether the <see cref="P:System.Windows.Data.Binding.Source"></see> property should be persisted.</summary>
        /// <returns>true if the property value has changed from its default; otherwise, false.</returns>
        [EditorBrowsable(EditorBrowsableState.Never)]
        public bool ShouldSerializeSource()
        {
            return binding.ShouldSerializeSource();
        }
        /// <summary>Indicates whether the <see cref="P:System.Windows.Data.Binding.ValidationRules"></see> property should be persisted.</summary>
        /// <returns>true if the property value has changed from its default; otherwise, false.</returns>
        [EditorBrowsable(EditorBrowsableState.Never)]
        public bool ShouldSerializeValidationRules()
        {
            return binding.ShouldSerializeValidationRules();
        }
        /// <summary>
        /// This member is not is not intended to be used directly from your code.
        /// </summary>
        [DefaultValue((string)null)]
        public object AsyncState
        {
            get { return binding.AsyncState; }
            set { binding.AsyncState = value; }
        }
        /// <summary>Gets or sets a value that indicates whether to evaluate the <see cref="P:System.Windows.Data.Binding.Path"></see> relative to the data item or the <see cref="T:System.Windows.Data.DataSourceProvider"></see> object.</summary>
        /// <returns>false to evaluate the path relative to the data item itself; otherwise, true. The default value is false.</returns>
        [DefaultValue(false)]
        public bool BindsDirectlyToSource
        {
            get { return binding.BindsDirectlyToSource; }
            set { binding.BindsDirectlyToSource = value; }
        }
        /// <summary>Gets or sets the converter to use.</summary>
        /// <returns>A value of type <see cref="T:System.Windows.Data.IValueConverter"></see>. The default value is null.</returns>
        [DefaultValue((string)null)]
        public IValueConverter Converter
        {
            get { return binding.Converter; }
            set { binding.Converter = value; }
        }
        /// <summary>Gets or sets the culture in which to evaluate the converter.</summary>
        /// <returns>The default value is null.</returns>
        [DefaultValue((string)null), TypeConverter(typeof(CultureInfoIetfLanguageTagConverter))]
        public CultureInfo ConverterCulture
        {
            get { return binding.ConverterCulture; }
            set { binding.ConverterCulture = value; }
        }
        /// <summary>Gets or sets the parameter to pass to the <see cref="P:System.Windows.Data.Binding.Converter"></see>.</summary>
        /// <returns>The parameter to pass to the <see cref="P:System.Windows.Data.Binding.Converter"></see>. The default value is null.</returns>
        [DefaultValue((string)null)]
        public object ConverterParameter
        {
            get { return binding.ConverterParameter; }
            set { binding.ConverterParameter = value; }
        }
        /// <summary>Gets or sets the name of the element to use as the binding source object.</summary>
        /// <returns>The value of the Name property or x:Name Attribute of the element of interest. You can refer only to elements that are created in code if they are registered to the appropriate <see cref="T:System.Windows.NameScope"></see> through RegisterName. For more information, see WPF Namescopes.The default value is null.</returns>
        [DefaultValue((string)null)]
        public string ElementName
        {
            get { return binding.ElementName; }
            set { binding.ElementName = value; }
        }
        /// <summary>Gets or sets a value that indicates whether the <see cref="T:System.Windows.Data.Binding"></see> should get and set values asynchronously.</summary>
        /// <returns>The default value is null.</returns>
        [DefaultValue(false)]
        public bool IsAsync
        {
            get { return binding.IsAsync; }
            set { binding.IsAsync = value; }
        }
        /// <summary>Gets or sets a value that indicates the direction of the data flow in the binding.</summary>
        /// <returns>One of the <see cref="T:System.Windows.Data.BindingMode"></see> values. The default value is <see cref="F:System.Windows.Data.BindingMode.Default"></see>, which returns the default binding mode value of the target dependency property. However, the default value varies for each dependency property. In general, user-editable control properties, such as those of text boxes and check boxes, default to two-way bindings, whereas most other properties default to one-way bindings.A programmatic way to determine whether a dependency property binds one-way or two-way by default is to get the property metadata of the property using <see cref="M:System.Windows.DependencyProperty.GetMetadata(System.Type)"></see> and then check the Boolean value of the <see cref="P:System.Windows.FrameworkPropertyMetadata.BindsTwoWayByDefault"></see> property.</returns>
        [DefaultValue(4)]
        public BindingMode Mode
        {
            get { return binding.Mode; }
            set { binding.Mode = value; }
        }
        /// <summary>Gets or sets a value that indicates whether to raise the <see cref="E:System.Windows.Data.Binding.SourceUpdated"></see> event when a value is transferred from the binding target to the binding source.</summary>
        /// <returns>true if the <see cref="E:System.Windows.Data.Binding.SourceUpdated"></see> event should be raised when the binding source value is updated; otherwise, false. The default value is false.</returns>
        [DefaultValue(false)]
        public bool NotifyOnSourceUpdated
        {
            get { return binding.NotifyOnSourceUpdated; }
            set { binding.NotifyOnSourceUpdated = value; }
        }
        /// <summary>Gets or sets a value that indicates whether to raise the <see cref="E:System.Windows.Data.Binding.TargetUpdated"></see> event when a value is transferred from the binding source to the binding target.</summary>
        /// <returns>true if the <see cref="E:System.Windows.Data.Binding.TargetUpdated"></see> event should be raised when the binding target value is updated; otherwise, false. The default value is false.</returns>
        [DefaultValue(false)]
        public bool NotifyOnTargetUpdated
        {
            get { return binding.NotifyOnTargetUpdated; }
            set { binding.NotifyOnTargetUpdated = value; }
        }
        /// <summary>Gets or sets a value that indicates whether to raise the <see cref="E:System.Windows.Controls.Validation.Error"></see> attached event on the bound object.</summary>
        /// <returns>true if the <see cref="E:System.Windows.Controls.Validation.Error"></see> attached event should be raised on the bound object when there is a validation error during source updates; otherwise, false. The default value is false.</returns>
        [DefaultValue(false)]
        public bool NotifyOnValidationError
        {
            get { return binding.NotifyOnValidationError; }
            set { binding.NotifyOnValidationError = value; }
        }
        /// <summary>Gets or sets the path to the binding source property.</summary>
        /// <returns>The path to the binding source. The default value is null.</returns>
        public PropertyPath Path
        {
            get { return binding.Path; }
            set { binding.Path = value; }
        }
        /// <summary>Gets or sets the binding source by specifying its location relative to the position of the binding target.</summary>
        /// <returns>A <see cref="T:System.Windows.Data.RelativeSource"></see> object specifying the relative location of the binding source to use. The default value is null.</returns>
        [DefaultValue((string)null)]
        public RelativeSource RelativeSource
        {
            get { return binding.RelativeSource; }
            set { binding.RelativeSource = value; }
        }
        /// <summary>Gets or sets the object to use as the binding source.</summary>
        /// <returns>The object to use as the binding source.</returns>
        public object Source
        {
            get { return binding.Source; }
            set { binding.Source = value; }
        }
        /// <summary>Gets or sets a handler you can use to provide custom logic for handling exceptions that the binding engine encounters during the update of the binding source value. This is only applicable if you have associated an <see cref="T:System.Windows.Controls.ExceptionValidationRule"></see> with your binding.</summary>
        /// <returns>A method that provides custom logic for handling exceptions that the binding engine encounters during the update of the binding source value.</returns>
        [DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
        public UpdateSourceExceptionFilterCallback UpdateSourceExceptionFilter
        {
            get { return binding.UpdateSourceExceptionFilter; }
            set { binding.UpdateSourceExceptionFilter = value; }
        }
        /// <summary>Gets or sets a value that determines the timing of binding source updates.</summary>
        /// <returns>One of the <see cref="T:System.Windows.Data.UpdateSourceTrigger"></see> values. The default value is <see cref="F:System.Windows.Data.UpdateSourceTrigger.Default"></see>, which returns the default <see cref="T:System.Windows.Data.UpdateSourceTrigger"></see> value of the target dependency property. However, the default value for most dependency properties is <see cref="F:System.Windows.Data.UpdateSourceTrigger.PropertyChanged"></see>, while the <see cref="P:System.Windows.Controls.TextBox.Text"></see> property has a default value of <see cref="F:System.Windows.Data.UpdateSourceTrigger.LostFocus"></see>.A programmatic way to determine the default <see cref="P:System.Windows.Data.Binding.UpdateSourceTrigger"></see> value of a dependency property is to get the property metadata of the property using <see cref="M:System.Windows.DependencyProperty.GetMetadata(System.Type)"></see> and then check the value of the <see cref="P:System.Windows.FrameworkPropertyMetadata.DefaultUpdateSourceTrigger"></see> property.</returns>
        [DefaultValue(0)]
        public UpdateSourceTrigger UpdateSourceTrigger
        {
            get { return binding.UpdateSourceTrigger; }
            set { binding.UpdateSourceTrigger = value; }
        }
        /// <summary>Gets a collection of rules that check the validity of the user input.</summary>
        /// <returns>A collection of <see cref="T:System.Windows.Controls.ValidationRule"></see> objects.</returns>
        public Collection<ValidationRule> ValidationRules
        {
            get { return binding.ValidationRules; }
        }

        public bool ValidatesOnDataErrors
        {
            get { return binding.ValidatesOnDataErrors; }
            set { binding.ValidatesOnDataErrors = value; }
        }

        public bool ValidatesOnExceptions
        {
            get { return binding.ValidatesOnExceptions; }
            set { binding.ValidatesOnExceptions = value; }
        }

        /// <summary>Gets or sets an XPath query that returns the value on the XML binding source to use.</summary>
        /// <returns>The XPath query. The default value is null.</returns>
        [DefaultValue((string)null)]
        public string XPath
        {
            get { return binding.XPath; }
            set { binding.XPath = value; }
        }

        #endregion
    }
}