﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.ComponentModel.DataAnnotations;
using System.Linq;
using System.Linq.Expressions;
using System.Reflection;

namespace Samples.Validation
{
	public partial class ValidationModelView : INotifyPropertyChanged, IDataErrorInfo
	{
		private static readonly Dictionary<string, Func<ValidationModelView, object>>
			propertyGetters = typeof(ValidationModelView).GetProperties()
							  .Where(p => GetValidations(p).Length != 0)
							  .ToDictionary(p => p.Name, p => GetValueGetter(p));

		private static readonly Dictionary<string, ValidationAttribute[]> validators =
			typeof(ValidationModelView).GetProperties()
			.Where(p => GetValidations(p).Length != 0)
			.ToDictionary(p => p.Name, p => GetValidations(p));

		private int range;
		private string required;
		private string noValidation;

		public event PropertyChangedEventHandler PropertyChanged;

		[Required(ErrorMessage = "Field 'Range' is required.")]
		[Range(1, 10, ErrorMessage = "Field 'Range' is out of range.")]
		public int Range
		{
			get
			{
				return this.range;
			}
			set
			{
				if (this.range != value)
				{
					this.range = value;
					this.OnPropertyChanged("Range");
				}
			}
		}

		[Required(ErrorMessage = "Field 'Required' is required.")]
		[StringLength(5, ErrorMessage = "Field 'Requires' must be smaller then 5 characters.")]
		public string Required
		{
			get
			{
				return this.required;
			}
			set
			{
				if (this.required != value)
				{
					this.required = value;
					this.OnPropertyChanged("Required");
				}
			}
		}

		public string NoValidation
		{
			get
			{
				return this.noValidation;
			}
			set
			{
				if (this.noValidation != value)
				{
					this.noValidation = value;
					this.OnPropertyChanged("NoValidation");
				}
			}
		}

		public string Error
		{
			get
			{
				var errors = from i in validators
							 from v in i.Value
							 where !v.IsValid(propertyGetters[i.Key](this))
							 select v.ErrorMessage;
				return string.Join(Environment.NewLine, errors.ToArray());
			}
		}

		public string this[string columnName]
		{
			get
			{
				if (propertyGetters.ContainsKey(columnName))
				{
					var value = propertyGetters[columnName](this);
					var errors = validators[columnName].Where(v => !v.IsValid(value))
						.Select(v => v.ErrorMessage).ToArray();
					this.OnPropertyChanged("Error");
					return string.Join(Environment.NewLine, errors);
				}

				this.OnPropertyChanged("Error");
				return string.Empty;
			}
		}

		private static ValidationAttribute[] GetValidations(PropertyInfo property)
		{
			return (ValidationAttribute[])property.GetCustomAttributes(typeof(ValidationAttribute), true);
		}

		private static Func<ValidationModelView, object> GetValueGetter(PropertyInfo property)
		{
			var instance = Expression.Parameter(typeof(ValidationModelView), "i");
			var cast = Expression.TypeAs(Expression.Property(instance, property), typeof(object));
			return (Func<ValidationModelView, object>)Expression.Lambda(cast, instance).Compile();
		}

		private void OnPropertyChanged(string propertyName)
		{
			if (this.PropertyChanged != null)
			{
				this.PropertyChanged(this, new PropertyChangedEventArgs(propertyName));
			}
		}
	}
}